#ifndef _DEBUG_H
#define _DEBUG_H

#include <stdio.h>
#ifdef SYSLOG
#include <syslog.h>
#endif

#ifdef SYSLOG
# define LOG(level, str, args...) do { printf(str, ## args); \
				printf("\n"); \
				syslog(level, str, ## args); } while(0)
# define OPEN_LOG(name) openlog(name, 0, 0)
#define CLOSE_LOG() closelog()
#else
# define LOG_EMERG	"EMERGENCY!"
# define LOG_ALERT	"ALERT!"
# define LOG_CRIT	"critical!"
# define LOG_WARNING	"warning"
# define LOG_ERR	"error"
# define LOG_INFO	"info"
# define LOG_DEBUG	"debug"
//# define LOG(level, str, args...) do { printf("%s, ", level); printf(str, ## args); printf("\n"); } while(0)

# define OPEN_LOG(name) do {;} while(0)
# define LOG(...) do {;} while(0)
#define CLOSE_LOG() do {;} while(0)
#endif

#if !(defined(DHCPD_DEBUG) || defined(DHCPC_DEBUG))

#ifdef DEBUG
# undef DEBUG
# define DEBUG(level, str, args...) LOG(level, str, ## args)
# define DEBUGGING
#else
# define DEBUG(level, str, args...) do {;} while(0)
#endif

#else //DHCPD_DEBUG

extern int enable_debug;
extern const char *dbgfile;
extern int __debug(const char* func, int line, const char* str, ... );

#define DEBUG(l, fmt, args...) do { \
	if (enable_debug) \
		__debug(__FUNCTION__, __LINE__, fmt, ##args); \
} while(0)

#undef LOG
#define LOG DEBUG

#endif //DHCPD_DEBUG

#define IPSTR "%u.%u.%u.%u"
#define IP2STR(addr) \
	((unsigned char *)&(addr))[0], \
	((unsigned char *)&(addr))[1], \
	((unsigned char *)&(addr))[2], \
	((unsigned char *)&(addr))[3]

#define MACSTR "%02X:%02X:%02X:%02X:%02X:%02X"
#define MAC2STR(m) \
	((unsigned char*)m)[0], ((unsigned char*)m)[1], ((unsigned char*)m)[2], \
	((unsigned char*)m)[3], ((unsigned char*)m)[4], ((unsigned char*)m)[5]

extern const char *dhcp_statestr(int state);
extern const char *dhcp_msgstr(int msgtype);

#endif
